/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.backend;

import java.util.Objects;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;

import com.redhat.thermostat.common.utils.LoggingUtils;

/**
 * An asynchronous backend activator, designed to process activate/deactivate
 * calls such that any misbehaving backends will not block the other backends
 * from activating or deactivating. This class is designed to be thread safe.
 * Once no more activation and deactivation will be done, it should have the
 * {@link #shutdown()} method called. Any operations requested from this object
 * after it has shutdown will do nothing (and will print a warning message).
 *
 * Note: This implementation currently does not handle removing all backends,
 * but not shutting down the agent itself.
 */
public class BackendActivator {

    private static final Logger logger = LoggingUtils.getLogger(BackendActivator.class);
    private static final int NO_SCHEDULE_DELAY_SECONDS = 0;
    private static final int AMOUNT_OF_RESUBMISSIONS_TO_EXECUTOR = 4;
    private static final int RESUBMISSION_DURATION_SECONDS = 15;
    private static final int STARTING_BACKEND_TRACKER_ATTEMPT_NUMBER = 1;

    private final long timeoutSeconds;
    private final ScheduledExecutorService activationExecutorService;
    private final ScheduledExecutorService deactivationExecutorService;
    private final ScheduledExecutorService timeoutExecutorService;
    private final ExecutorService postActivateService;
    private final CountDownLatch allBackendsDeactivated;
    private AtomicBoolean shutdown = new AtomicBoolean(false);
    private AtomicInteger numBackendsActivate = new AtomicInteger(0);

    /**
     * Creates a new backend activator with the provided number of threads and
     * timeout to (attempt to) kill misbehaving backends. This will generate
     * a total of (2 * numBackendsInParallel) + 2 threads.
     * @param numBackendsInParallel How many threads this should allocate to
     *                              activating/deactivating backends.
     * @param timeoutSeconds How many seconds before a backend is considered to
     *                       be misbehaving and should attempt to interrupt it.
     *                       This should be non-negative.
     * @throws IllegalArgumentException If the timeout is negative.
     */
    public BackendActivator(int numBackendsInParallel, long timeoutSeconds) {
        this(timeoutSeconds, Executors.newScheduledThreadPool(numBackendsInParallel),
                Executors.newScheduledThreadPool(numBackendsInParallel), Executors.newScheduledThreadPool(1),
                Executors.newFixedThreadPool(1), new CountDownLatch(1));
    }

    BackendActivator(long timeoutSeconds, ScheduledExecutorService activationExecutorService,
                     ScheduledExecutorService deactivationExecutorService, ScheduledExecutorService timeoutExecutorService,
                     ExecutorService postActivateService, CountDownLatch allBackendsDeactivated) {
        if (timeoutSeconds < 0) {
            throw new IllegalArgumentException("Cannot have a negative backend activation timeout");
        }

        this.timeoutSeconds = timeoutSeconds;
        this.activationExecutorService = Objects.requireNonNull(activationExecutorService);
        this.deactivationExecutorService = Objects.requireNonNull(deactivationExecutorService);
        this.timeoutExecutorService = Objects.requireNonNull(timeoutExecutorService);
        this.postActivateService = Objects.requireNonNull(postActivateService);
        this.allBackendsDeactivated = Objects.requireNonNull(allBackendsDeactivated);
    }

    /**
     * Keeps recursively scheduling the backend tracker every period of
     * RESUBMISSION_DURATION_SECONDS. Termination of recursion occurs when the
     * attempt number is greater than AMOUNT_OF_RESUBMISSIONS_TO_EXECUTOR.
     */
    private void scheduleBackendTrackerRecursively(final BackendTracker backendTracker) {
        timeoutExecutorService.schedule(new Runnable() {
            @Override
            public void run() {
                if (!backendTracker.getFuture().isDone() && backendTracker.getAttemptNumber() <= AMOUNT_OF_RESUBMISSIONS_TO_EXECUTOR) {
                    String logMessage = String.format("Backend '%s' is taking a while to activate/deactivate [%d/%d]",
                                                      backendTracker.getBackend(), backendTracker.getAttemptNumber(),
                                                      AMOUNT_OF_RESUBMISSIONS_TO_EXECUTOR);
                    logger.warning(logMessage);
                    backendTracker.incrementAttemptNumber();
                    scheduleBackendTrackerRecursively(backendTracker);
                }
            }
        }, RESUBMISSION_DURATION_SECONDS, TimeUnit.SECONDS);
    }

    private void scheduleAction(final Backend backend, final boolean isActivating, Runnable runnable) {
        ScheduledExecutorService executor = (isActivating ? activationExecutorService : deactivationExecutorService);
        final ScheduledFuture<?> future = executor.schedule(runnable, NO_SCHEDULE_DELAY_SECONDS, TimeUnit.SECONDS);

        timeoutExecutorService.schedule(new Runnable() {
            @Override
            public void run() {
                if (!future.isDone()) {
                    logger.warning("Backend '" + backend + "' is taking too long to " + (isActivating ? "activate" : "deactivate"));
                    future.cancel(true);
                }
            }
        }, timeoutSeconds, TimeUnit.SECONDS);

        BackendTracker backendTracker = new BackendTracker(backend, future, STARTING_BACKEND_TRACKER_ATTEMPT_NUMBER);
        scheduleBackendTrackerRecursively(backendTracker);
    }

    /**
     * Activates a backend. Will attempt to kill activation if it takes longer
     * than the value set in the constructor for timing out.
     * @param backend The backend to activate, should not be null.
     * @param postActivate Any actions to be performed after activation. This
     *                     may be null if you do not want to run anything after
     *                     activation.
     * @throws NullPointerException If the backend is null.
     */
    public void activateBackend(final Backend backend, final Runnable postActivate) {
        Objects.requireNonNull(backend);

        if (shutdown.get()) {
            logger.info("Cannot activate '" + backend + "', backend activator shut down");
            return;
        }

        logger.info("Adding backend: " + backend);
        scheduleAction(backend, true, new Runnable() {
            @Override
            public void run() {
                final boolean didActivate = backend.activate();
                if (!didActivate) {
                    logger.warning("Backend '" + backend + "' failed to activate.");
                }
                if (postActivate != null && didActivate) {
                    postActivateService.submit(postActivate);
                }
                numBackendsActivate.incrementAndGet();
            }
        });
    }

    /**
     * Deactivates a backend. Will attempt to kill it if it takes longer than
     * the value set in the constructor for timing out.
     * @param backend The backend to activate, should not be null.
     * @param postDeactivate Any actions to be performed after activation. This
     *                       may be null if you do not want to run anything
     *                       after deactivation.
     * @throws NullPointerException If the backend is null.
     */
    public void deactivateBackend(final Backend backend, final Runnable postDeactivate) {
        Objects.requireNonNull(backend);

        if (shutdown.get()) {
            logger.info("Cannot deactivate '" + backend + "', backend activator shut down");
            return;
        }

        logger.info("Removing backend: " + backend);
        scheduleAction(backend, false, new Runnable() {
            @Override
            public void run() {
                if (!backend.deactivate()) {
                    logger.warning("Backend '" + backend + "' failed to deactivate.");
                    return;
                }
                if (postDeactivate != null) {
                    postActivateService.submit(postDeactivate);
                }
                if (numBackendsActivate.decrementAndGet() == 0) {
                    allBackendsDeactivated.countDown();
                }
            }
        });
    }

    /**
     * Waits for backends to deactivate. This depends on how many backends you
     * have registered with this object, as it counts the number activated and
     * deactivated, and will either return when the net change of activation to
     * deactivation is zero, or if it times out based on the provided argument.
     * @param timeoutSeconds How long until this method will return. If this is
     *                       zero of negative, it will return immediately as
     *                       stated in the CountDownLatch specification.
     * @return True if there was no timeout, false otherwise.
     */
    public boolean waitForAllBackendsToDeactivate(int timeoutSeconds) {
        try {
            if (allBackendsDeactivated.await(timeoutSeconds, TimeUnit.SECONDS)) {
                return true;
            } else {
                logger.warning("Timed out waiting for backends to deactivate");
            }
        } catch (InterruptedException e) {
            logger.warning("Waiting for deactivation latch interrupted");
        }

        return false;
    }

    /**
     * Shuts down the backend activator so it can no longer accept any backends
     * for activation or deactivation.
     */
    public void shutdown() {
        if (shutdown.get()) {
            logger.warning("Already shut down the backend activator");
            return;
        }

        activationExecutorService.shutdownNow();
        deactivationExecutorService.shutdownNow();
        timeoutExecutorService.shutdownNow();
        postActivateService.shutdownNow();

        shutdown.set(true);
    }

    // For unit testing.
    int getNumberOfBackendsActive() {
        return numBackendsActivate.get();
    }

    static class BackendTracker {
        private Backend backend;
        private Future<?> future;
        private int attemptNumber;

        BackendTracker(Backend backend, Future<?> future, int attemptNumber) {
            this.backend = Objects.requireNonNull(backend);
            this.future = Objects.requireNonNull(future);
            this.attemptNumber = attemptNumber;
        }

        void incrementAttemptNumber() {
            attemptNumber++;
        }

        Backend getBackend() {
            return backend;
        }

        Future<?> getFuture() {
            return future;
        }

        int getAttemptNumber() {
            return attemptNumber;
        }
    }
}
