/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.vm.classstat.agent.internal;


import static org.mockito.Matchers.anyListOf;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.net.URI;
import java.util.Arrays;


import com.redhat.thermostat.agent.http.HttpRequestService;
import com.redhat.thermostat.common.config.experimental.ConfigurationInfoSource;
import com.redhat.thermostat.common.plugin.BackendConfigurationUtil;
import com.redhat.thermostat.common.plugin.SystemID;
import com.redhat.thermostat.vm.classstat.model.VmClassStat;

import org.junit.Before;
import org.junit.Test;

public class VmClassStatDAOTest {

    private static final String VM_ID = "jvmId";
    private static final Long LOADED_CLASSES = 12345L;
    private static final Long LOADED_BYTES = 2345L;
    private static final Long UNLOADED_CLASSES = 3456L;
    private static final Long UNLOADED_BYTES = 4567L;
    private static final Long CLASS_LOAD_TIME = 5678L;

    private static final String WRITER_ID = "writerId";
    private static final long TIMESTAMP = 123455667L;
    private static final String SOME_SYSTEM_ID = "somesystemid";
    private static final String AGENT_ID = "some-agent";
    private static final String JSON = "{\"this\":\"is\",\"also\":\"JSON\"}";

    private static final URI GATEWAY_URI;

    static {
        GATEWAY_URI = URI.create("http://example.com/jvm-classtat/");
    }

    private VmClassStat classStat;
    private VmClassStatDAOImpl.JsonHelper jsonHelper;
    private VmClassStatDAOImpl dao;
    private HttpRequestService httpRequestService;

    @Before
    public void setUp() throws IOException {
        this.classStat = new VmClassStat(WRITER_ID,
                VM_ID,
                TIMESTAMP,
                LOADED_CLASSES,
                LOADED_BYTES,
                UNLOADED_CLASSES,
                UNLOADED_BYTES,
                CLASS_LOAD_TIME
        );

        jsonHelper = mock(VmClassStatDAOImpl.JsonHelper.class);
        when(jsonHelper.toJson(anyListOf(VmClassStat.class))).thenReturn(JSON);

        ConfigurationInfoSource source = mock(ConfigurationInfoSource.class);
        BackendConfigurationUtil config = mock(BackendConfigurationUtil.class);
        when(config.getGatewayURL()).thenReturn(GATEWAY_URI);
        VmClassStatDAOImpl.ConfigurationCreator creator = mock(VmClassStatDAOImpl.ConfigurationCreator.class);
        when(creator.create(source)).thenReturn(config);

        httpRequestService = mock(HttpRequestService.class);
        dao = new VmClassStatDAOImpl(jsonHelper, creator);
        dao.bindHttpRequestService(httpRequestService);
        dao.bindConfigurationInfoSource(source);
    }

    @Test
    public void verifyPut() throws Exception {
        SystemID id = mock(SystemID.class);
        when(id.getSystemID()).thenReturn(SOME_SYSTEM_ID);
        dao.bindSystemID(id);
        dao.activate();
        dao.put(classStat);

        verify(jsonHelper).toJson(eq(Arrays.asList(classStat)));
        String s = verify(httpRequestService).sendHttpRequest(JSON, GATEWAY_URI.resolve("systems/" + SOME_SYSTEM_ID + "/jvms/" + VM_ID), HttpRequestService.Method.POST);
    }
}

